/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  libgrg_utils.c - utility functions for libgringotts
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "config.h"
#include "libgringotts.h"
#include "libgrg_crypt.h"
#include "libgrg_structs.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <ctype.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>

#include <mcrypt.h>
#include <mhash.h>

// Basically, the following 2 functions are the adaptation of string.h
// functions for byte sequences. Strings are byte sequence that are
// \0-terminated; in this case, this is dangerous, because a \0 can
// occour _inside_ the very sequence. Of course, there's no way to tell
// how long the sequence is, other than "remembering" it, and passing it
// to the functions.

/**
 * grg_memdup
 * @src: the source
 * @len: its length
 *
 * Duplicates a byte sequence into a new one
 *
 * Returns: a newly allocated (to free afterwards) byte sequence
 */
unsigned char *
grg_memdup (const unsigned char *src, unsigned long len)
{
	unsigned char *ret;
	
	if(!src || !len)
		return NULL;

	ret = (unsigned char *) malloc (len);
	memcpy (ret, src, len);

	return ret;
}

/**
 * grg_memconcat:
 * @count: the number of byte sequences to concatenate
 * @src1: the first byte sequence to concatenate
 * @len1: the length of the first sequence
 * @...: the other sequences/lengths
 *
 * Concatenates some byte sequences
 *
 * Returns: a newly allocated (to free afterwards) byte sequence
 */
unsigned char *
grg_memconcat (const int count, ...)
{
	va_list ap;
	unsigned char *ret, *tmp, *strings[count];
	int dim[count], i, dimtot = 0;

	if (!count)
		return NULL;

	va_start (ap, count);

	for(i=0; i<count; i++){
		strings[i] = va_arg(ap, unsigned char *);
		dim[i] = va_arg(ap,int);
		dimtot += dim[i];
	}

	if (!dimtot)
		return NULL;

	ret = (unsigned char *) malloc (dimtot);

	tmp = ret;
	for(i=0; i<count; i++){
		memcpy (tmp, strings[i], dim[i]);
		tmp += dim[i];
	}

	va_end (ap);

	return ret;
}

/**
 * grg_XOR_mem:
 * @src: the byte sequence to XOR
 * @src_len: its length
 * @mask: the byte sequence to XOR the first with. It will be repeated if shorter.
 * @mask_length: its length
 *
 * XORs a byte sequence with another, eventually repeating the latter to match the first.
 */
void
grg_XOR_mem (unsigned char *src, int src_len, unsigned char *mask, int mask_len)
{
	int i;
	for (i=0; i<src_len; i++)
		src[i] ^= mask[i % mask_len];
}	

/**
 * grg_long2char:
 * @seed: the long to convert
 *
 * Converts a long into four bytes
 *
 * Returns: a newly allocated 4-bytes sequence, to free after use
 */
unsigned char *
grg_long2char (const unsigned long seed)
{
	unsigned char *ret;
	long tmp = seed;
	int i;

	ret = (unsigned char *) malloc (4);

	for (i = 3; i >= 0; i--, tmp >>= 8)
		ret[i] = tmp & 0x0ff;

	return ret;
}

/**
 * grg_char2long:
 * @seed: the 4-char sequence to convert
 *
 * Reverts grg_long2char(), converting back into a long
 *
 * Returns: a long
 */
unsigned long
grg_char2long (const unsigned char *seed)
{
	//Many thanks to James Antill (james@and.org) for the patch!!
	unsigned long ret = 0;
	//unsigned const char *true_seed_what_a_shame_on_me_thanks_James = seed;
	int i;

	for (i = 3; i >= 0; i--)
		ret |= /*true_seed_what_a_shame_on_me_thanks_James*/seed [i] << ((3 - i) * 8);

	return ret;
}

/**
 * grg_rnd_seq:
 * @size: the size of the sequence to generate
 *
 * Generates a random sequence of bytes.
 *
 * Returns: a newly-allocated byte sequence
 */
unsigned char *
grg_rnd_seq (const GRG_CTX gctx, const unsigned int size)
{
	unsigned char *ret = (unsigned char *) malloc (size);

	if (gctx->rnd > 2)
	{
		read (gctx->rnd, ret, size);
		return ret;
	}
	else
	{
		int i;

		for (i = 0; i < size; i++)
			ret[i] = rand () % 256;
		return ret;
	}
}

/**
 * grg_rnd_chr:
 *
 * Returns a random byte.
 *
 * Returns: a random byte.
 */
unsigned char
grg_rnd_chr (const GRG_CTX gctx)
{
	if (gctx->rnd > 2)
	{
		unsigned char rnd;
		read (gctx->rnd, &rnd, 1);
		return rnd;
	}
	else
		return (unsigned char)(rand () * 256);
}

/**
 * grg_free:
 * @pntr: pointer to the memory to free
 * @dim: length of the sequence; if -1 it must be NULL-terminated
 *
 * Frees a sequence of bytes, overwriting it with random data
 */
void
grg_free (const GRG_CTX gctx, void *alloc_data, const long dim)
{
	unsigned int i = 0;
	unsigned char *pntr = (char *) alloc_data;

	if (!pntr)
		return;

	if(gctx->sec_lvl > GRG_SEC_MIN)
	{
		if (dim < 0)
			while (pntr[i])
				pntr[i++] = grg_rnd_chr (gctx);
		else
		{
			unsigned char *rnd = grg_rnd_seq (gctx, dim);
			memcpy (pntr, rnd, dim);
			free (rnd);
		}
	}

	free (pntr);
}

/**
 * grg_ascii_pwd_quality:
 * @pwd: a non-multibyte string with the password
 * @pwd_len: the maximum length, to grant a termination. If
 * negative, the string _must_ be null-terminated
 *
 * Calculates an indicative value proportional to the password
 * "quality". Returns a value from 0 to 1, where 0 is the worst
 * password (4 char, all numeric), 1 the approximation of 256
 * bits of real data in the password, or better. The scale is 
 * logarythmic, so it will grow faster for shorter passwords than 
 * for larger ones, since improvements in a scarcely-secure password
 * are more valuable than improvements in better ones.<br>
 * Anyway, this should be used only for ASCII-like passwords, null
 * terminated. If you want to eval more "generic" sequences of 
 * bytes, use grg_file_pwd_quality. If you use multibyte encodings,
 * you must convert them to your locale encoding before evaluation.
 * 
 * Returns: a double between 0 and 1, inclusive
 */
double
grg_ascii_pwd_quality (const unsigned char *pwd, const unsigned long pwd_len)
{
	int A=FALSE, a=FALSE, n=FALSE, p=FALSE;
	long i=0;
	long tmp = (pwd_len < 0) ? LONG_MAX-1 : pwd_len;
	int basin = 0;
	double ret;

	if(!pwd)
		return 0.0;
	
	while((pwd[i] != '\0') && (i<tmp)){

		if(islower(pwd[i])){
			a=TRUE;
			i++;
			continue;
		}

		if(isdigit(pwd[i])){
			n=TRUE;
			i++;
			continue;
		}

		if(isupper(pwd[i])){
			A=TRUE;
			i++;
			continue;
		}
		
		p=TRUE;
		i++;
	}
	
	if(i<4)
		return 0.0;
	
	if(a)
		basin += 26;
	if(A)
		basin += 26;
	if(n)
		basin += 10;
	if(p)
		basin += 32;

	#undef LOG2
	#define LOG2(val) \
		(log(val)/0.693147) 	/* log, basis 2 */
	#define QVAL(length, num_char) \
		(length*LOG2(num_char))	/* number of `real' bits in pwd */
	#define QMAX	5.545177	//log(256)
	#define QMIN	2.586840	//log(4*LOG2(10))
	#define QDIF	2.958337	//QMAX-QMIN
	
	ret = (log(QVAL(i, basin))-QMIN)/QDIF;
	
	#undef LOG2
	#undef QVAL
	#undef QMAX
	#undef QMIN
	#undef QDIF

	if (ret < 0)
		return 0.0;
	if (ret > 1)
		return 1.0;
	return ret;
}

/**
 * grg_file_pwd_quality:
 * @pwd_path: the path to the file to use as password
 *
 * Calculates an indicative value proportional to the password
 * "quality", in the case you want to use a file content as a
 * password. Returns a value from 0 to 1, where 0 is the worst
 * case (empty file), 1 the approximation of 256 bits of real 
 * data, or better. The scale is linear.
 * 
 * Returns: a double between 0 and 1, inclusive
 */
double
grg_file_pwd_quality (const unsigned char *pwd_path)
{
	double ret;
	int pdf;

	pdf = open(pwd_path, O_RDONLY);

	if (pdf < 3)
	{
		close (pdf);
		return 0.0;
	}

	ret = ((double)(lseek(pdf, 0, SEEK_END))) / 32.0;

	close (pdf);

	if (ret < 0)
		return 0.0;
	if (ret > 1)
		return 1.0;
	return ret;
}
