/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  grg_entries.c - functions to manage the GList of Gringotts' entries
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <stdlib.h>
#include <string.h>

#include "../libgringotts/libgringotts.h"

#include "grg_defs.h"
#include "gringotts.h"
#include "grg_prefs.h"
#include "grg_safe.h"
#include "grg_widgets.h"
#include "grg_recent_dox.h"

struct grg_entry
{
	gchar *entryID;
	gchar *entryBody;
};

static GList *entries = NULL;
static GList *current = NULL;
static guchar *serialized;
static gint pos_to_restore;
static gboolean newer_data = FALSE;

/**
 * meta_free:
 * @data: the callback's data
 * @user_data: the callback's user-defined data
 *
 * Frees a single node. Used only by grg_entries_free()
 */
static void
meta_free (gpointer data, gpointer user_data)
{
	struct grg_entry *entry = (struct grg_entry *) data;

	grg_free (gctx, entry->entryID, -1);
	entry->entryID = NULL;
	grg_free (gctx, entry->entryBody, -1);
	entry->entryBody = NULL;
	g_free (entry);
	entry = NULL;
}

/**
 * grg_entries_append:
 * @ID: the title of the new entry
 *
 * Appends a new entry to the list
 */
void
grg_entries_append (void)
{
	struct grg_entry *entry;

	entry = g_malloc (sizeof (struct grg_entry));

	entry->entryID = g_strdup ("");
	entry->entryBody = g_strdup ("");

	entries = g_list_append (entries, entry);
	current = g_list_last (entries);
}

/**
 * grg_entries_remove:
 *
 * Removes the current entry and frees it
 */
void
grg_entries_remove (void)
{
	gint pos;

	if (!current)
		return;

	pos = g_list_position (entries, current);
	meta_free (current->data, NULL);
	entries = g_list_remove_link (entries, current);
	g_list_free_1 (current);
	current = g_list_nth (entries, pos);
	if (!current)
		current = g_list_nth (entries, pos - 1);
}

/**
 * grg_entries_is_first:
 *
 * Tells if the current entry is the first one
 *
 * Returns: TRUE if true, FALSE if false.
 */
gboolean
grg_entries_is_first (void)
{
	if (!current)
		return TRUE;
	if (!g_list_previous (current))
		return TRUE;
	return FALSE;
}

/**
 * grg_entries_is_last:
 *
 * Tells if the current entry is the last one
 *
 * Returns: TRUE if true, FALSE if false.
 */
gboolean
grg_entries_is_last (void)
{
	if (!current)
		return TRUE;

	if (!g_list_next (current))
		return TRUE;

	return FALSE;
}

/**
 * grg_entries_is_empty:
 *
 * Tells if the current list is empty
 *
 * Returns: TRUE if true, FALSE if false.
 */
gboolean
grg_entries_is_empty (void)
{
	return (entries == NULL);
}

/**
 * grg_entries_first:
 *
 * Goes to the first entry
 */
void
grg_entries_first (void)
{
	current = g_list_first (entries);
}

/**
 * grg_entries_prev:
 *
 * Goes to the previous entry
 */
void
grg_entries_prev (void)
{
	if (!grg_entries_is_first ())
		current = g_list_previous (current);
}

/**
 * grg_entries_next:
 *
 * Goes to the next entry
 */
void
grg_entries_next (void)
{
	if (!grg_entries_is_last ())
		current = g_list_next (current);
}

/**
 * grg_entries_last:
 *
 * Goes to the last entry
 */
void
grg_entries_last (void)
{
	current = g_list_last (entries);
}

/**
 * grg_entries_nth:
 * @pos: the position to go to
 *
 * Goes to the specified entry, if possible
 */
void
grg_entries_nth (gint pos)
{
	current = g_list_nth (entries, pos);
}

/**
 * grg_entries_position:
 *
 * Tells the position of the current entry in the list
 *
 * Returns: a gint with the position
 */
gint
grg_entries_position (void)
{
	if (!current)
		return -1;

	return g_list_position (entries, current);
}

/**
 * grg_entries_raise
 *
 * Shifts one position up
 */
void
grg_entries_raise (void)
{
	gpointer ent;

	if (!current->prev)
		return;

	ent = current->data;
	current->data = current->prev->data;
	current->prev->data = ent;
}

/**
 * grg_entries_sink
 *
 * Shifts one position down
 */
void
grg_entries_sink (void)
{
	gpointer ent;

	if (!current->next)
		return;

	ent = current->data;
	current->data = current->next->data;
	current->next->data = ent;
}

/**
 * grg_entries_get_ID
 *
 * Returns the title of the current entry.
 *
 * Returns: a gchar* with the title (NOT a copy), or NULL if the list is empty
 */
gchar *
grg_entries_get_ID (void)
{
	if (current)
		return ((struct grg_entry *) current->data)->entryID;
	return NULL;
}

/**
 * grg_entries_get_Body
 *
 * Returns the body text of the current entry.
 *
 * Returns: a gchar* with the text, or NULL if the list is empty
 */
gchar *
grg_entries_get_Body (void)
{
	if (current)
		return ((struct grg_entry *) current->data)->entryBody;
	return NULL;
}

static void
grg_entries_set_ID_asis (gchar * ID)
{
	if(current)
	{
		grg_free (gctx, ((struct grg_entry *) current->data)->entryID, strlen (((struct grg_entry *) current->data)->entryID));
		((struct grg_entry *) current->data)->entryID = ID;
	}else
		grg_free (gctx, ID, -1);
}

/**
 * grg_entries_set_ID
 * @ID: the text to store
 *
 * Stores the given title in the current entry.
 */
void
grg_entries_set_ID (const gchar * ID)
{
	if(ID)
		grg_entries_set_ID_asis (g_strdup(ID));
}

static void
grg_entries_set_Body_asis (gchar * Body)
{
	if(current)
	{
		grg_free (gctx, ((struct grg_entry *) current->data)->entryBody, strlen (((struct grg_entry *) current->data)->entryBody));
		((struct grg_entry *) current->data)->entryBody = Body;
	}
	else
		grg_free (gctx, Body, -1);
}

/**
 * grg_entries_set_Body
 * @Body: the text to store
 *
 * Stores the given text as the body of the current entry.
 */
void
grg_entries_set_Body (const gchar * Body)
{
	if (Body)
		grg_entries_set_Body_asis (g_strdup (Body));
}

/**
 * grg_entries_free:
 *
 * Deletes and frees all the list
 */
void
grg_entries_free (void)
{
	g_list_foreach (entries, meta_free, NULL);
	g_list_free (entries);
	entries = NULL;
	current = NULL;
}

/**
 * meta_save:
 * @data: the callback's data
 * @user_data: the callback's user-defined data
 *
 * "serializes" a single node. Used only by grg_entries_save()
 */
static void
meta_save (gpointer data, gpointer user_data)
{
	struct grg_entry *entry = (struct grg_entry *) data;
	gchar *eBody, *eID, *res;
	gint dim;
	
	dim = strlen(entry->entryBody);
	eBody = g_markup_escape_text (entry->entryBody, dim);

	dim = strlen(entry->entryID);
	eID = g_markup_escape_text (entry->entryID, dim);
	
#define XML_TOP	"\n<entry>\n<title>"
#define XML_MID "</title>\n<body>"
#define XML_BOT "</body>\n</entry>"

	res = g_strconcat (serialized, XML_TOP, eID, XML_MID, eBody, XML_BOT, "", NULL);
	
	grg_free (gctx, eBody, -1);
	eBody = NULL;
	grg_free (gctx, eID, -1);
	eID = NULL;

	grg_free (gctx, serialized, -1);
	serialized = g_strdup (res);

	grg_free (gctx, res, -1);
	res = NULL;
}

static gchar *
g_markup_unescape_text (const gchar *txt, const gint len)
{
	gint i, cnt=0;
	gchar *ret = (gchar *) malloc (len);

	for(i=0; i<len; i++, cnt++)
		if(txt[i] != '&')
			ret[cnt] = txt[i];
		else
			switch (txt [i+1]){
				case 'q':
					ret[cnt] = '"';
					i += 5;
					break;
				case 'l':
					ret[cnt] = '<';
					i += 3;
					break;
				case 'g':
					ret[cnt] = '>';
					i += 3;
					break;
				default:
					if(txt [i+2] == 'm'){
						ret[cnt] = '&';
						i += 4;
					}else{
						ret[cnt] = '\'';
						i += 5;
					}
			}

	ret = (gchar *) realloc (ret, cnt+1);	
	ret[cnt] = '\0';

	return ret;
}

/**
 * grg_entries_save:
 * @file: the path of the file to save
 * @key: the keyring to use
 *
 * Saves the list into an encrypted file
 *
 * Returns: GRG_OK if all is well, an error if not
 */
gint
grg_entries_save (gchar * file, GRG_KEY key)
{
	gint err, pos = grg_entries_position();

	serialized = g_strdup_printf
		("<save_file_fmt_version>\n" GRG_FILE_SUBVERSION "\n</save_file_fmt_version>"
		 "\n<position>\n%d</position>", pos);
	
	g_list_foreach (entries, meta_save, NULL);

	err = grg_encrypt_file (gctx, key, file, serialized, strlen (serialized));

	grg_free (gctx, serialized, -1);
	serialized = NULL;

	if (err != GRG_OK)
		return err;

	grg_recent_dox_push (file);
	return GRG_OK;
}

//the letter denoting the field is the first of the tag text
#define	VERS_FIELD		's'
#define TITLE_FIELD		't'
#define BODY_FIELD		'b'
#define POSITION_FIELD	'p'
#define IGNORE_FIELD	'e'

static void
get_tag (GMarkupParseContext *context,
				const gchar	*element_name,
				const gchar	**attribute_names,
				const gchar	**attribute_values,
				gpointer user_data,
				GError	**error)
{
	*((gchar *) user_data) = element_name[0];
	/* more formally 
	switch (element_name[0]){
		case 't':
			*((gchar *) user_data) = TITLE_FIELD;
			break;
		case 'b':
			*((gchar *) user_data) = BODY_FIELD;
			break;
		default:
	}*/
}

static void
trash_tag (GMarkupParseContext *context,
				const gchar *element_name,
				gpointer user_data,
				GError **error)
{
	*((gchar *) user_data) = IGNORE_FIELD;
}

static void
compose_entry (GMarkupParseContext *context,
				const gchar *text,
				gsize text_len,  
				gpointer user_data,
				GError **error)
{
	gchar *tmp = g_markup_unescape_text (text, text_len);
	switch (*((gchar *) user_data)){
		case TITLE_FIELD:
			grg_entries_append ();
			grg_entries_set_ID_asis (tmp);
			break;
		case BODY_FIELD:
			grg_entries_set_Body_asis (tmp);
			break;
		case POSITION_FIELD:
			pos_to_restore = atoi (text);
			break;
		case VERS_FIELD:
			newer_data = (text[0] > GRG_FILE_SUBVERSION[0]);
			break;
		default:
	}
}

/**
 * grg_entries_load_from_string_v3:
 * @str: the string which contains the data, file format 3
 *
 * "de-serializes" a string into an entry list
 */
static void
grg_entries_load_from_string_v3 (gchar *str)
{
	gint end;
	gchar field;
	GMarkupParser *context = (GMarkupParser *) g_malloc (sizeof (GMarkupParser));
	GMarkupParseContext *parser;
	GError *err = NULL;

	context->start_element = get_tag;
	context->end_element = trash_tag;
	context->text = compose_entry;
	context->passthrough = NULL;
	context->error = NULL;

	parser = g_markup_parse_context_new (context, 0, (gpointer) &field, NULL);
	end = strlen(str);

	g_markup_parse_context_parse (parser, str, end, &err);
	if(!err)
		g_markup_parse_context_end_parse (parser, &err);

	g_markup_parse_context_free (parser);
	if(err)
		g_error_free (err);
	g_free (context);

	grg_entries_nth (pos_to_restore);
}

/**
 * grg_entries_load_from_string_v2:
 * @str: the string which contains the data, file format 2
 *
 * "de-serializes" a string into an entry list
 */
static void
grg_entries_load_from_string_v2 (gchar * str)
{
	gchar **fields;
	gint num;
	
	fields = g_strsplit (str, "\n", 0);

	num = 0;
	while (fields[num * 2] != NULL)
	{
		gchar *Body;

		grg_entries_append ();
		grg_entries_set_ID (fields[num * 2]);
		Body = g_strcompress (fields[num * 2 + 1]);
		grg_entries_set_Body (Body);

		grg_free (gctx, Body, -1);
		Body = NULL;
		grg_free (gctx, fields[num*2], -1);
		grg_free (gctx, fields[num*2+1], -1);
		num++;
	}

	grg_entries_first ();

	g_free (fields);
}

#define V2_TOKEN 		"<save_file_fmt_version>\n"
#define V2_TOKEN_LEN	24

void
grg_entries_load_from_string (gchar * str, GtkWidget *parent)
{
	grg_entries_free ();

	if (strncmp(str, V2_TOKEN, V2_TOKEN_LEN) != 0)
		grg_entries_load_from_string_v2 (str);
	else
		grg_entries_load_from_string_v3 (str);
	if (newer_data)
		grg_msg (_("The file has been created with a newer version of Gringotts. If you save it using this one, some formatting can be lost."), GTK_MESSAGE_WARNING, parent);
}

/**
 * grg_load_wrapper:
 * @txt: a pointer to a byte sequence to store the data in. It must be freed after use!
 * @pwd: the password to decode data
 * @file: path of the file to read data from
 * @just_validate: if TRUE, it only checks for file validity, without loading it.
 * 
 * Wrapper to grg_load_crypted, to add UTF-8 validation.
 *
 * Returns: 0 if OK; an error code otherwise (see grg_crypt.h)
 */
gint
grg_load_wrapper (guchar ** txt, GRG_KEY key, const guchar * file,
		  gboolean just_validate)
{
	gint err;
	gulong len = 0;

	if (just_validate)
		err = grg_validate_file (gctx, file);
	else
		err = grg_decrypt_file (gctx, key, file, txt, &len);

	grg_prefs_update();

	if ((err != GRG_OK) || just_validate)
		return err;

	if (just_validate)
		return GRG_OK;

	if (!g_utf8_validate (*txt, -1, NULL))
	{
		grg_free (gctx, serialized, -1);
		serialized = NULL;
		return GRG_READ_INVALID_CHARSET_ERR;
	}

	grg_recent_dox_push (file);
	return GRG_OK;
}

/**
 * grg_entries_n_el:
 *
 * Tells how many elements does the list contain
 *
 * Returns: a guint with the elements' number
 */
guint
grg_entries_n_el (void)
{
	if (entries)
		return g_list_length (entries);
	else
		return 0;
}

/**
 * grg_entries_find:
 * @needle: the text to find
 * @offset: the offset to search from
 * @only_current: if TRUE, consider only the current entry
 * @case_sens: wheter to be case sensitive or not
 *
 * Searches for a text in the entries.
 *
 * Returns: the offset of the found text, -1 if not found
 */
glong
grg_entries_find (gchar * needle, glong offset, gboolean only_current,
		  gboolean case_sens)
{
	guchar *text = grg_entries_get_Body (), *start = g_utf8_offset_to_pointer (text, offset), *occur;
	glong result;

	if (case_sens)
	{
		occur = g_strstr_len (start, strlen (start), needle);

		if (occur)
			result = g_utf8_pointer_to_offset (text,
							   occur) + offset;
		else
			result = -1;
	}
	else
	{
		gchar *tok1, *tok2;
		gulong len;

		tok1 = g_utf8_casefold (start, -1);
		tok2 = g_utf8_casefold (needle, -1);
		len = strlen (tok1);

		occur = g_strstr_len (tok1, len, tok2);

		if (occur)
			result = g_utf8_pointer_to_offset (tok1,
							   occur) + offset;
		else
			result = -1;

		grg_free (gctx, tok1, len);
		tok1 = NULL;
		g_free (tok2);
		tok2 = NULL;
	}

	if (result < 0)
	{
		gint cur = grg_entries_position (), ret;

		if (only_current)
			return -1;

		if (grg_entries_is_last ())
			return -1;
		else
			grg_entries_next ();

		ret = grg_entries_find (needle, 0, FALSE, case_sens);
		if (ret == -1)
			grg_entries_nth (cur);
		return ret;
	}

	return result;
}
